from collections import namedtuple
from typing import Optional, Tuple

import colorcorrect.algorithm as cca
import numpy as np
from colorcorrect.util import from_pil, to_pil
from PIL import Image, ImageChops, ImageEnhance, ImageOps


class Retouch:
    """
    Perform retouching.

    Passing an image and options to the constructor will convert the specified image.
    """

    RGB_RED_VALUE: int = 255
    RGB_GREEN_VALUE: int = 255
    RGB_BLUE_VALUE: int = 255
    RGB_WHITE_COLOR: Tuple[int, int, int] = (
        RGB_RED_VALUE,
        RGB_GREEN_VALUE,
        RGB_BLUE_VALUE,
    )

    def __init__(
        self,
        image: Image,
        colorautoadjust: bool = False,
        colorstretch: bool = False,
        grayscale: bool = False,
        invert: bool = False,
        tosaka: Optional[float] = None,
        outputrgb: bool = False,
        noise: Optional[float] = None,
    ) -> None:
        """
        __init__ constructor.

        Args:
            image (Image): Image file.
            colorautoadjust (bool, optional): Use colorautoadjust algorithm. Defaults to False.
            colorstretch (bool, optional): Use colorstretch algorithm. Defaults to False.
            grayscale (bool, optional): Convert to grayscale. Defaults to False.
            invert (bool, optional): Invert color. Defaults to False.
            tosaka (Optional[float], optional): Use Tosaka mode. Defaults to None.
            outputrgb (bool, optional): Outputs a monochrome image in RGB. Defaults to False.
            noise (Optional[float], optional): Add Gaussian noise. Defaults to None.
        """
        self.image = image
        self.colorautoadjust = colorautoadjust
        self.colorstretch = colorstretch
        self.grayscale = grayscale
        self.invert = invert
        self.tosaka = tosaka
        self.output_rgb = outputrgb
        self.noise = noise

    def process(self) -> Image:
        self.image = self.__rgba_convert()

        if self.invert:
            self.image = self.__invert()

        if self.colorautoadjust:
            self.image = self.__colorautoadjust()

        if self.colorstretch:
            self.image = self.__colorstretch()

        if self.grayscale:
            self.image = self.__grayscale()

        if self.noise is not None:
            self.image = self.__noise()

        if self.tosaka is not None:
            self.image = self.__tosaka()

        if self.output_rgb:
            self.image = self.__output_rgb()

        return self.image

    def __colorautoadjust(self) -> Image:
        """
        __colorautoadjust

        Use colorautoadjust algorithm.

        Returns:
            Image: processed image.
        """
        if self.image.mode == "L":
            return self.image
        return to_pil(cca.automatic_color_equalization(from_pil(self.image)))

    def __colorstretch(self) -> Image:
        """
        __colorstretch

        Use colorstretch algorithm.

        Returns:
            Image: processed image.
        """
        if self.image.mode == "L":
            return self.image
        return to_pil(cca.stretch(cca.grey_world(from_pil(self.image))))

    def __noise(self) -> Image:
        """
        __noise

        Add Gaussian noise.
        To add noise, you need to specify floating-point numbers;
        a value of about 10.0 will be just right.

        Returns:
            Image: processed image.
        """
        table = [x * 2 for x in range(256)] * len(self.image.getbands())
        if self.image.mode == "RGB":
            noise_image = Image.effect_noise(
                (self.image.width, self.image.height), self.noise
            ).convert("RGB")
            self.image = ImageChops.multiply(self.image, noise_image).point(table)
        if self.image.mode == "L":
            noise_image = Image.effect_noise(
                (self.image.width, self.image.height), self.noise
            )
            self.image = ImageChops.multiply(self.image, noise_image).point(table)
        return self.image

    def __grayscale(self) -> Image:
        """
        __grayscale

        Converts to grayscale based on the luminance in the CIE XYZ color space.

        Returns:
            Image: processed image.
        """
        if self.image.mode == "L":
            return self.image

        rgb = np.array(self.image, dtype="float32")

        GAMMA = 2.2
        ColorMatrix = namedtuple("ColorMatrix", ("red", "green", "blue"))

        # Coefficients for luminance in CIE XYZ color space
        CIE_XYZ = ColorMatrix(0.2126, 0.7152, 0.0722)

        # Inverse Gamma Correction
        rgbL = pow(rgb / 255.0, GAMMA)

        # Extract RGB values
        r, g, b = rgbL[:, :, 0], rgbL[:, :, 1], rgbL[:, :, 2]

        # Convert to grayscale
        grayL = CIE_XYZ.red * r + CIE_XYZ.green * g + CIE_XYZ.blue * b

        # gamma correction
        gray = pow(grayL, 1.0 / GAMMA) * 255

        self.image = Image.fromarray(gray.astype("uint8"))

        return self.image

    def __invert(self) -> Image:
        """
        __invert

        Invert color.

        Returns:
            Image: processed image.
        """
        return ImageOps.invert(self.image)

    def __tosaka(self) -> Image:
        """
        __tosaka

        Use Tosaka mode.

        Tosaka mode is a mode that expresses the preference of
        Tosaka-senpai, a character in "Kyūkyoku Chōjin R",
        for "photos taken with Tri-X that look like they were
        burned onto No. 4 or No. 5 photographic paper".
        Only use floating-point numbers when using this mode;
        numbers around 2.4 will make it look right.

        Returns:
            Image: processed image.
        """
        if self.image.mode != "L":
            self.image = self.__grayscale()
        imageC = ImageEnhance.Contrast(self.image)
        self.image = imageC.enhance(self.tosaka)
        return self.image

    def __rgba_convert(self) -> Image:
        """
        __rgba_convert

        Converts image data that contains transparency to image data that does not contain transparency.

        Returns:
            Image: processed image.
        """
        if self.image.mode == "RGBA":
            self.image.load()
            background = Image.new("RGB", self.image.size, self.RGB_WHITE_COLOR)
            background.paste(self.image, mask=self.image.split()[3])
            self.image = background
        if self.image.mode == "LA":
            self.image.load()
            background = Image.new("L", self.image.size, 255)
            background.paste(self.image, mask=self.image.split()[1])
            self.image = background
        return self.image

    def __output_rgb(self) -> Image:
        """
        __output_rgb

        Outputs a monochrome image in RGB.

        Returns:
            Image: processed image.
        """
        if self.image.mode == "L":
            self.image = self.image.convert("RGB")
        return self.image



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['anshitsu']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.3.1,<10.0.0',
 'colorcorrect>=0.9.1,<0.10.0',
 'fire>=0.4.0,<0.5.0',
 'numpy>=1.21.0,<2.0.0']

entry_points = \
{'console_scripts': ['anshitsu = anshitsu.cli:main']}

setup_kwargs = {
    'name': 'anshitsu',
    'version': '1.4.0',
    'description': 'A tiny digital photographic utility.',
    'long_description': '# Anshitsu\n\n[![Testing](https://github.com/huideyeren/anshitsu/actions/workflows/testing.yml/badge.svg)](https://github.com/huideyeren/anshitsu/actions/workflows/testing.yml)\n\n[![codecov](https://codecov.io/gh/huideyeren/anshitsu/branch/main/graph/badge.svg?token=ZYRX8NBTLQ)](https://codecov.io/gh/huideyeren/anshitsu)\n\nA tiny digital photographic utility.\n\n"Anshitsu" means a darkroom in Japanese.\n\n## Install\n\nRun this command in an environment where Python 3.8 or higher is installed.\n\nWe have tested it on Windows, Mac, and Ubuntu on GitHub Actions, but we have not tested it on Macs with Apple Silicon, so please use it at your own risk on Macs with Apple Silicon.\n\n``` shell\npip install anshitsu\n```\n\n## Usage\n\nIt is as described in the following help.\n\n``` shell\nINFO: Showing help with the command \'anshitsu -- --help\'.\n\nNAME\n    anshitsu - Process Runnner for Command Line Interface\n\nSYNOPSIS\n    anshitsu PATH <flags>\n\nDESCRIPTION\n    This utility converts the colors of images such as photos.\n\n    If you specify a directory path, it will convert\n    the image files in the specified directory.\n    If you specify a file path, it will convert the specified file.\n    If you specify an option, the specified conversion will be performed.\n\n    Tosaka mode is a mode that expresses the preference of\n    Tosaka-senpai, a character in "Kyūkyoku Chōjin R",\n    for "photos taken with Tri-X that look like they were\n    burned onto No. 4 or No. 5 photographic paper".\n    Only use floating-point numbers when using this mode;\n    numbers around 2.4 will make it look right.\n\nPOSITIONAL ARGUMENTS\n    PATH\n        Type: str\n        Directory or File Path\n\nFLAGS\n    --colorautoadjust=COLORAUTOADJUST\n        Type: bool\n        Default: False\n        Use colorautoadjust algorithm. Defaults to False.\n    --colorstretch=COLORSTRETCH\n        Type: bool\n        Default: False\n        Use colorstretch algorithm. Defaults to False.\n    --grayscale=GRAYSCALE\n        Type: bool\n        Default: False\n        Convert to grayscale. Defaults to False.\n    --invert=INVERT\n        Type: bool\n        Default: False\n        Invert color. Defaults to False.\n    --tosaka=TOSAKA\n        Type: Optional[Optional]\n        Default: None\n        Use Tosaka mode. Defaults to None.\n    --outputrgb=OUTPUTRGB\n        Type: bool\n        Default: False\n        Outputs a monochrome image in RGB. Defaults to False.\n    --noise=NOISE\n        Type: Optional[Optional]\n        Default: None\n        Add Gaussian noise. Defaults to None.\n\nNOTES\n    You can also use flags syntax for POSITIONAL ARGUMENTS\n```\n\nIf a directory is specified in the path, an `out` directory will be created in the specified directory, and the converted JPEG and PNG images will be stored in JPEG format.\n\nIf you specify a JPEG or PNG image file as the path, an `out` directory will be created in the directory where the image is stored, and the converted image will be stored in JPEG format.\n\nNote: If you specify\nIf you specify a file of any other format in the path, error handling is not available. An error will probably occur and the program will terminate abnormally.\n\n## Algorithm\n\nThe following algorithms are available in this tool.\n\n### RGBA to RGB Convert\n\nConverts an image that contains Alpha, such as RGBA, to image data that does not contain Alpha.\nTransparent areas will be filled with white.\n\nThis algorithm is performed on any image file.\n\n### invert\n\nInverts the colors of an image using Pillow\'s built-in algorithm.\n\nIn the case of negative film, color conversion that takes into account the film base color is not performed, but we plan to follow up with a feature to be developed in the future.\n\n### colorautoajust\n\nWe will use the "automatic color equalization" algorithm described in the following paper to apply color correction.\n\nThis process is more time consuming than the algorithm used in "colorstretch", but it can reproduce more natural colors.\n\n(References)\n\nA. Rizzi, C. Gatta and D. Marini, "A new algorithm for unsupervised global and local color correction.", Pattern Recognition Letters, vol. 24, no. 11, 2003.\n\n### colorstretch\n\nThe "gray world" and "stretch" algorithms described in the following paper are combined to apply color correction.\n\nThis process is faster than the algorithm used in "colorautoajust".\n\n(References)\n\nD. Nikitenko, M. Wirth and K. Trudel, "Applicability Of White-Balancing Algorithms to Restoring Faded Colour Slides: An Empirical Evaluation.", Journal of Multimedia, vol. 3, no. 5, 2008.\n\n### grayscale\n\nConvert a color image to grayscale using the algorithm described in the following article.\n\n[Python でグレースケール(grayscale)化](https://qiita.com/yoya/items/dba7c40b31f832e9bc2a#pilpillow-%E3%81%A7%E3%82%B0%E3%83%AC%E3%83%BC%E3%82%B9%E3%82%B1%E3%83%BC%E3%83%AB%E5%8C%96-numpy-%E3%81%A7%E4%BD%8E%E8%BC%9D%E5%BA%A6%E5%AF%BE%E5%BF%9C)\n\nNote: This article is written in Japanese.\n\n### Tosaka mode\n\nTosaka mode is a mode that expresses the preference of Tosaka-senpai, a character in "Kyūkyoku Chōjin R", for "photos taken with Tri-X that look like they were burned onto No. 4 or No. 5 photographic paper".\n\nOnly use floating-point numbers when using this mode; numbers around 2.4 will make it look right.\n\nWhen this mode is specified, color images will also be converted to grayscale.\n\n### outputrgb\n\nOutputs a monochrome image in RGB.\n\n### noise\n\nAdd Gaussian noise.\n\nTo add noise, you need to specify floating-point numbers; a value of about 10.0 will be just right.\n\n## Special Thanks\n\nWe are using the following libraries.\n\n[shunsukeaihara/colorcorrect](https://github.com/shunsukeaihara/colorcorrect)\n',
    'author': 'Iosif Takakura',
    'author_email': 'iosif@huideyeren.info',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/huideyeren',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
